@@ -9,6 +9,7 @@ import pandas
 import warnings
 from random import seed
 from copy import deepcopy
+from datetime import datetime
 
 import cobra
 import symengine
@@ -68,8 +69,9 @@ def save_output(riptide_obj='NULL', path='NULL', file_type='JSON', silent=False)
     if path == 'NULL':
         if silent == False:
             print('WARNING: Did not provide an output directory. Using default riptide_files in working directory')
-        path = 'riptide_files'
+        path = riptide_obj.model.id + '_' + riptide_obj.run_start
     try:
+        path = path + '_' + riptide_obj.run_start
         os.mkdir(path)
     except:
         if silent == False:
@@ -341,10 +343,10 @@ def _rarefy(abunds, n):
 
 
 # Simplified to run in parallel
-def multi_contextualize(args, frac):
+def _multi_contextualize(args, frac):
     current_fit = contextualize(model=args['model'], transcriptome=args['transcriptome'], fraction=frac, samples=args['samples'], 
                          minimum=args['minimum'], conservative=args['conservative'], objective=args['objective'], 
-                         additive=args['additive'], set_bounds=args['set_bounds'], 
+                         additive=args['additive'], set_bounds=args['set_bounds'], task_frac=args['task_frac'],
                          tasks=args['tasks'], exclude=args['exclude'], gpr=args['gpr'], threshold=args['threshold'], 
                          phase=2, silent=True)
 
@@ -357,36 +359,41 @@ def _find_best_fit(frac_range, argDict, prev_best=None):
     progress = 0.0
 
     if prev_best == None:
+        best_fit_concordance = 0.
         if argDict['silent'] == False:
             print('Analyzing context-specific subnetwork flux ranges...')
     else:
-        best_fit = prev_best
+        best_fit_concordance = prev_best.concordance
         if argDict['silent'] == False:
-            print('Testing local objective fractions to ' + str(best_fit.fraction_of_optimum) + '...')
+            print('Testing local objective fractions to ' + str(prev_best.fraction_of_optimum) + '...')
 
     if argDict['silent'] == False: sys.stdout.write('\rProgress: 0%')
 
     maxfit_report = {}
-    best_fit = multi_contextualize(argDict, frac_range[0])
+    best_fit = _multi_contextualize(argDict, frac_range[0])
     maxfit_report[best_fit.fraction_of_optimum] = best_fit.concordance
     
     progress += increment
-    if isinstance(best_fit.concordance, str) != True:
-        best_fit_concordance = 0.
-        improved = 0
-        if argDict['silent'] == False: sys.stdout.write('\rProgress: ' + str(float("%.2f" % progress)) + '%        ')
+    improved = 0
+    best_fit_concordance = 0.
+    if isinstance(best_fit.concordance, str) == True:
+        if argDict['silent'] == False: sys.stdout.write('\rProgress: ' + str(float("%.2f" % progress)) + '%')
+    elif best_fit.concordance['r'] > best_fit_concordance:
+        best_fit_concordance = best_fit.concordance['r']
+        improved += 1
+        if argDict['silent'] == False: sys.stdout.write('\rProgress: ' + str(float("%.2f" % progress)) + '%  -  improved fit identified (' + str(improved) + ') ')
     else:
-        improved = 1
-        if argDict['silent'] == False: sys.stdout.write('\rProgress: ' + str(float("%.2f" % progress)) + '%  -  fit identified        ')
-    
+        if argDict['silent'] == False: sys.stdout.write('\rProgress: ' + str(float("%.2f" % progress)) + '%')
+        
     # Identify best fit of flux sample to transcriptome
-    improvement = False
     for frac in frac_range[1:]:
+        improvement = False
         try:
-            fit = multi_contextualize(argDict, frac)
+            fit = _multi_contextualize(argDict, frac)
         except:
             progress += increment
-            if argDict['silent'] == False: sys.stdout.write('\rProgress: ' + str(float("%.2f" % progress)) + '%        ')
+            if argDict['silent'] == False:
+                sys.stdout.write('\rProgress: ' + str(float("%.2f" % progress)) + '% ')
             continue
             
         maxfit_report[fit.fraction_of_optimum] = fit.concordance
@@ -398,23 +405,39 @@ def _find_best_fit(frac_range, argDict, prev_best=None):
         
         progress += increment
         if argDict['silent'] == False:
-            if improvement == False:
-                sys.stdout.write('\rProgress: ' + str(float("%.2f" % progress)) + '%        ')
-            elif improved == 0 or prev_best == None:
+            if improvement == False and improved == 0:
+                sys.stdout.write('\rProgress: ' + str(float("%.2f" % progress)) + '% ')
+            elif improved == 0 and prev_best == None:
                 improved += 1
-                sys.stdout.write('\rProgress: ' + str(float("%.2f" % progress)) + '%  -  fit identified        ')
+                sys.stdout.write('\rProgress: ' + str(float("%.2f" % progress)) + '%  -  fit identified                     ')
             else:
-                sys.stdout.write('\rProgress: ' + str(float("%.2f" % progress)) + '%  -  improved fit identified        ')
+                improved += 1
+                sys.stdout.write('\rProgress: ' + str(float("%.2f" % progress)) + '%  -  improved fit identified (' + str(improved) + ') ')
 
     if argDict['silent'] == False:
-        if improvement == False:
-            sys.stdout.write('\rProgress: 100%                          \n\n')
+        if improvement == False and improved == 0:
+            sys.stdout.write('\rProgress: 100% \n\n')
+            sys.stdout.flush()
+        elif improved == 0 and prev_best == None:
+            sys.stdout.write('\rProgress: 100%  -  fit identified                       \n\n')
             sys.stdout.flush()
-        elif improved == 0 or prev_best == None:
-            sys.stdout.write('\rProgress: ' + str(float("%.2f" % progress)) + '%  -  fit identified        \n\n')
         else:
-            sys.stdout.write('\rProgress: 100%  -  improved fit identified        \n\n')
+            improved += 1
+            sys.stdout.write('\rProgress: 100%  -  improved fit identified (' + str(improved) + ') \n\n')
             sys.stdout.flush()
+    
+    # Check if metabolic tasks can carry flux in the best fit model
+    if argDict['silent'] == False and len(argDict['tasks']) >= 1:
+        obj_ID = str(best_fit.model.objective.expression).split()[0].split('*')[-1]
+        failed = 0
+        for task in argDict['tasks']:
+            best_fit.model.objective = task
+            task_flux = best_fit.model.slim_optimize(error_value=0.)
+            if task_flux <= argDict['threshold']:
+                print('WARNING:', task, 'carries no flux in best fitting model')
+                failed += 1
+        best_fit.model.objective = obj_ID
+        if failed >= 1: print('\n')
 
     best_fit.maxfit_report = maxfit_report
 
@@ -424,7 +447,7 @@ def _find_best_fit(frac_range, argDict, prev_best=None):
 # Iteratively run RIPTiDe over a range of objective minimum fractions
 def maxfit(model, transcriptome = 'none', frac_min = 0.1, frac_max = 0.9, frac_step = 0.1, prune = True,
     samples = 1000, minimum = False, conservative = False, objective = True, additive = False, 
-    set_bounds = True, silent = False, tasks = [], task_frac = 0.01, exclude = [], gpr = False, threshold = 1e-6):
+    set_bounds = True, silent = False, tasks = [], task_frac = 0.01, exclude = [], gpr = False, threshold = 1e-8):
 
     '''
     Iterative RIPTiDe for a range of minimum objective fluxes, returns model with best correlation 
@@ -479,7 +502,7 @@ def maxfit(model, transcriptome = 'none', frac_min = 0.1, frac_max = 0.9, frac_s
     tasks : list
         List of gene or reaction ID strings for forced inclusion in final model (metabolic tasks or essential genes)
     task_frac : float
-        Minimum fraction of optimal flux for metabolic task reactions during pruning
+        Minimum fraction of optimal flux for metabolic task reactions given preceeding constraints
         Default is 0.01
     exclude : list
         List of reaction ID strings for forced exclusion from final model
@@ -488,10 +511,11 @@ def maxfit(model, transcriptome = 'none', frac_min = 0.1, frac_max = 0.9, frac_s
         Default is False
     threshold : float
         Minimum flux a reaction must acheive in order to avoid pruning during flux sum minimization step
-        Default is 1e-5
+        Default is 1e-8
     '''
 
     iter_start = time.time()
+    curr_run = str(datetime.now()).replace(' ','_').replace('-','').replace(':','').split('.')[0].replace(' ','_')
     seed(937162211)
 
     if samples <= 100:
@@ -519,6 +543,14 @@ def maxfit(model, transcriptome = 'none', frac_min = 0.1, frac_max = 0.9, frac_s
             print('WARNING: Improper maximum fraction provided, setting to default value')
         frac_max = 0.85
 
+    if threshold+1e-6 < 1e-6:
+        threshold = abs(threshold)
+
+    if len(tasks) >= 1:
+        screened_tasks = _screen_tasks(model, tasks, silent)
+    if task_frac <= 0. or task_frac > 1.:
+        task_frac = 0.01
+
     frac_range = [round(x, 3) for x in list(numpy.arange(frac_min, frac_max+frac_step, frac_step))]
     if len(frac_range) == 1:
         if silent == False:
@@ -528,9 +560,9 @@ def maxfit(model, transcriptome = 'none', frac_min = 0.1, frac_max = 0.9, frac_s
         print('Running max fit RIPTiDe for objective fraction range:', frac_min, 'to', frac_max, '...')
 
     argDict = {'model':model, 'transcriptome':transcriptome, 'silent':silent, 
-               'samples':samples, 'minimum':minimum, 
+               'samples':samples, 'minimum':minimum, 'task_frac':task_frac,
                'conservative':conservative, 'objective':objective, 'additive':additive, 
-               'set_bounds':set_bounds, 'tasks':tasks, 'exclude':exclude, 
+               'set_bounds':set_bounds, 'tasks':screened_tasks, 'exclude':exclude, 
                'gpr':gpr, 'threshold':threshold, 'phase':2}
     
     warnings.filterwarnings('ignore', category=UserWarning)
@@ -543,12 +575,13 @@ def maxfit(model, transcriptome = 'none', frac_min = 0.1, frac_max = 0.9, frac_s
             frac_step = round(frac_step / 2.0, 3)
             small_frac_range += [round(top_fit.fraction_of_optimum - frac_step, 3), round(top_fit.fraction_of_optimum + frac_step, 3)]
 
-        new_fit = _find_best_fit(list(set(small_frac_range)), argDict, top_fit)
-        if isinstance(new_fit.concordance, str) != True:
-            if new_fit.concordance['r'] > top_fit.concordance['r']:
-                top_fit = new_fit
-                for frac in new_fit.maxfit_report.keys():
-                     all_maxfit[frac] = new_fit.maxfit_report[frac]
+        if len(set(small_frac_range)) >= 1:
+            new_fit = _find_best_fit(list(set(small_frac_range)), argDict, top_fit)
+            if isinstance(new_fit.concordance, str) != True:
+                if new_fit.concordance['r'] > top_fit.concordance['r']:
+                    top_fit = new_fit
+                    for frac in new_fit.maxfit_report.keys():
+                        all_maxfit[frac] = new_fit.maxfit_report[frac]
     
     warnings.filterwarnings('default', category=UserWarning)
     top_fit.maxfit_report = all_maxfit
@@ -567,6 +600,7 @@ def maxfit(model, transcriptome = 'none', frac_min = 0.1, frac_max = 0.9, frac_s
     report_dict = _operation_report(iter_start, model, top_fit.model, top_fit.concordance, silent, phase=1)
     top_fit.additional_parameters['operation'] = report_dict
     top_fit.maxfit = True
+    top_fit.run_start = curr_run
 
     return top_fit
 
@@ -602,7 +636,7 @@ def _screen_tasks(model, tasks, silent):
 # Create context-specific model based on transcript distribution
 def contextualize(model, transcriptome = 'none', samples = 1000, silent = False, prune = True,
     fraction = 0.8, minimum = False, conservative = False, objective = True, additive = False, direct = False,
-    set_bounds = True, tasks = [], task_frac = 0.01, exclude = [], gpr = False, threshold = 1e-6, phase=1):
+    set_bounds = True, tasks = [], task_frac = 0.01, exclude = [], gpr = False, threshold = 1e-8, phase=1):
 
     '''Reaction Inclusion by Parsimony and Transcriptomic Distribution or RIPTiDe
     
@@ -629,6 +663,8 @@ def contextualize(model, transcriptome = 'none', samples = 1000, silent = False,
     '''
 
     start_time = time.time()
+    curr_run = str(datetime.now()).replace(' ','_').replace('-','').replace(':','').split('.')[0]
+    
     seed(937162211)
 
     if _test_exchange_space(model, minimum=1e5):
@@ -642,6 +678,8 @@ def contextualize(model, transcriptome = 'none', samples = 1000, silent = False,
     riptide_object.additional_parameters['objective'] = objective
     riptide_object.additional_parameters['additive'] = additive
     riptide_object.additional_parameters['set_bounds'] = set_bounds
+    riptide_object.start_time = str(start_time)
+    riptide_object.run_start = curr_run
 
     # Correct some possible user error
     samples = int(samples)
@@ -652,10 +690,10 @@ def contextualize(model, transcriptome = 'none', samples = 1000, silent = False,
         fraction = 0.01
     elif fraction >= 1.0: 
         fraction = 0.99
-    if task_frac <= 0.0: 
-        task_frac = 0.01
-    elif task_frac >= 1.0: 
-        task_frac = 0.01
+    if threshold+1e-6 < 1e-6:
+        threshold = abs(threshold)
+    if task_frac <= 0. or task_frac > 1. or task_frac > fraction:
+        task_frac = fraction
     if minimum != False:
         if minimum <= 0.0: 
             minimum = 0.0001
@@ -705,7 +743,6 @@ def contextualize(model, transcriptome = 'none', samples = 1000, silent = False,
     
     # Define linear coefficients for both steps
     rxn_transcriptome, gene_hits = _transcript_to_reactions(transcriptome, model, gpr, additive)
-    rm_rxns = set([x.id for x in model.reactions])
     all_min_coefficient_dict = {}
     all_max_coefficient_dict = {}
     riptide_object.percent_of_mapping = gene_hits
@@ -713,6 +750,8 @@ def contextualize(model, transcriptome = 'none', samples = 1000, silent = False,
     if silent == False:
         if phase == 1:
             print('Pruning zero flux subnetworks...')
+            
+    inactive_rxns = set(exclude)
     for x in range(0, transcriptome['replicates']):
         current_rxn_transcriptome = {}
         for index in rxn_transcriptome.keys():
@@ -746,36 +785,30 @@ def contextualize(model, transcriptome = 'none', samples = 1000, silent = False,
             screened_tasks = _screen_tasks(model, tasks, silent)
             riptide_object.metabolic_tasks = screened_tasks
             for task in screened_tasks:
-                min_coefficient_dict[objective] = min_coefficient
-                max_coefficient_dict[objective] = max_coefficient
+                min_coefficient_dict[task] = min_coefficient
+                max_coefficient_dict[task] = max_coefficient
                 try:
-                    all_min_coefficient_dict[objective].append(min_coefficient_dict[objective])
-                    all_max_coefficient_dict[objective].append(max_coefficient_dict[objective])
+                    all_min_coefficient_dict[task].append(min_coefficient_dict[task])
+                    all_max_coefficient_dict[task].append(max_coefficient_dict[task])
                 except KeyError:
-                    all_min_coefficient_dict[objective] = [min_coefficient_dict[objective]]
-                    all_max_coefficient_dict[objective] = [max_coefficient_dict[objective]]
+                    all_min_coefficient_dict[task] = [min_coefficient_dict[task]]
+                    all_max_coefficient_dict[task] = [max_coefficient_dict[task]]
         else:
             screened_tasks = []
             
-        # Determine active network sections based on coefficients
-        active_rxns = _constrain_for_pruning(model=model, min_coefficients=min_coefficient_dict, fraction=fraction,
-                                             objective=objective, minimum_flux=minimum_threshold, silent=silent)
-        rm_rxns = rm_rxns.difference(active_rxns)
-        if len(screened_tasks) >= 1:
-            for task in screened_tasks:
-                active_rxns = _constrain_for_pruning(model=model, min_coefficients=min_coefficient_dict, fraction=task_frac,
-                                                     objective=task, minimum_flux=minimum_threshold, silent=silent)
-                rm_rxns = rm_rxns.difference(active_rxns)
-        riptide_object.minimization_coefficients = all_min_coefficient_dict
-                
+        # Determine inactive network sections based on coefficients
+        inactive_rxns |= _constrain_for_pruning(model=model, min_coefficients=min_coefficient_dict, 
+                                             objective=objective, obj_fraction=fraction, tasks=screened_tasks, task_fraction=task_frac, 
+                                             minimum_flux=minimum_threshold, silent=silent)
+    
     # Prune inactive model components
-    rm_rxns = rm_rxns.union(set(exclude))
-    if len(rm_rxns) > 0 and prune == True:
-        riptide_model = _prune_model(model, rm_rxns, conservative)
+    riptide_object.minimization_coefficients = all_min_coefficient_dict
+    if len(inactive_rxns) > 0 and prune == True:
+        riptide_model = _prune_model(model, inactive_rxns, conservative)
         riptide_object.pruned = _record_pruned_elements(model, riptide_model)
     else:
         riptide_model = deepcopy(model)
-
+         
     # Find optimal solution space based on transcription and final constraints
     if silent == False:
         if phase == 1:
@@ -788,7 +821,8 @@ def contextualize(model, transcriptome = 'none', samples = 1000, silent = False,
 
     # Sample weighted flux distributions
     flux_samples, concordance = _constrain_for_sampling(model=riptide_model, max_coefficients=max_coefficient_dict, sampling_depth=samples,
-                                                       objective=objective, frac=fraction, minimum_flux=minimum_threshold, silent=silent)
+                                                        objective=objective, obj_frac=fraction,  tasks=screened_tasks, task_frac=task_frac, 
+                                                        min_flux=minimum_threshold,silent=silent)
     riptide_object.flux_samples = flux_samples
     riptide_object.concordance = concordance
         
@@ -928,19 +962,19 @@ def _assign_coefficients(rxn_transcript_dict, model, direct):
 
 
 # Constrain task minimum flux
-def _constrain_task(model, task, frac, min_val, silent):
+def _constrain_task(model, task, frac, min_flux, silent):
     
     model.objective = task
     model.objective.direction = 'max'
     
     task_name = task + '_constraint'
-    task_bound1 = model.slim_optimize(error_value=0.)
-    if silent == False and task_bound1 <= min_val:
+    task_bound = model.slim_optimize(error_value=0.)
+    if silent == False and task_bound <= min_flux:
         print('WARNING:', task, 'minimum flux infeasible')
-        task_bounds = [model.reactions.get_by_id(task).lower_bound, model.reactions.get_by_id(task).upper_bound]
+        task_bounds = [min_flux, model.reactions.get_by_id(task).upper_bound]
     else:
-        task_bound2 = task_bound1 * frac
-        task_bounds = [task_bound1, task_bound2]
+        task_bounds = [task_bound, task_bound * frac]
+        
     task_expression = model.reactions.get_by_id(task).flux_expression
     task_constraint = model.problem.Constraint(task_expression, name=task_name, ub=max(task_bounds), lb=min(task_bounds))
     model.add_cons_vars(task_constraint)
@@ -968,40 +1002,50 @@ def _weighted_expression(model, coefficients={}):
 
 
 # Determine those reactions that carry flux in a pFBA objective set to a threshold of maximum
-def _constrain_for_pruning(model, min_coefficients, fraction, objective, minimum_flux, silent):
+def _constrain_for_pruning(model, min_coefficients, objective, obj_fraction, tasks, task_fraction, minimum_flux, silent):
     
     constrained_model = deepcopy(model)
     
     # Add objective/task constraints
+    min_coefficient = min(list(min_coefficients.values()))
     if isinstance(objective, str) == True:
-        min_coefficient = min(list(min_coefficients.values()))
         min_coefficients[objective] = min_coefficient
-        constrained_model = _constrain_task(constrained_model, objective, fraction, minimum_flux, silent)
-            
+        constrained_model = _constrain_task(constrained_model, objective, obj_fraction, minimum_flux, silent)
+
+    # metabolic task constraints
+    if len(tasks) >= 1:
+        for task in tasks:
+            min_coefficients[task] = min_coefficient
+            constrained_model = _constrain_task(constrained_model, task, task_fraction, minimum_flux, silent)
+    
     pruning_expr = _weighted_expression(model=constrained_model, coefficients=min_coefficients)
     constrained_model.objective = model.problem.Objective(pruning_expr, direction='min', sloppy=True)
     constrained_model.solver.update()
     
     constrained_fluxes = constrained_model.optimize().fluxes
-    active_rxns = set([rxn.id for rxn in constrained_model.reactions if abs(constrained_fluxes[rxn.id]) >= minimum_flux]) 
+    inactive_rxns = set([rxn.id for rxn in constrained_model.reactions if abs(constrained_fluxes[rxn.id]) == 0.]) 
     
-    return active_rxns
+    return inactive_rxns
 
 
 # Determine those reactions that carry flux in a pFBA objective set to a threshold of maximum
-def _constrain_for_sampling(model, max_coefficients, sampling_depth, objective, frac, minimum_flux, silent):
+def _constrain_for_sampling(model, max_coefficients, sampling_depth, objective, obj_frac, tasks, task_frac, min_flux, silent):
     
     constrained_model = deepcopy(model)
     
     # Apply weigths to new expression, allow deviation
     if isinstance(objective, str) == True:
-        constrained_model = _constrain_task(constrained_model, objective, frac, minimum_flux, silent)
+        constrained_model = _constrain_task(constrained_model, objective, obj_frac, min_flux, silent)
+    if len(tasks) >= 1:
+        for task in tasks:
+            constrained_model = _constrain_task(constrained_model, task, task_frac, min_flux, silent)
+
     sampling_expr = _weighted_expression(model=constrained_model, coefficients=max_coefficients)
     constrained_model.objective = model.problem.Objective(sampling_expr, direction='max', sloppy=True)
     constrained_model.solver.update()
     
     flux_sum = constrained_model.slim_optimize(error_value=0.)
-    flux_sum_constraint = constrained_model.problem.Constraint(sampling_expr, lb=flux_sum*0.01, ub=flux_sum)
+    flux_sum_constraint = constrained_model.problem.Constraint(sampling_expr, lb=min_flux, ub=flux_sum)
     model.add_cons_vars(flux_sum_constraint)
     constrained_model.solver.update()
     
@@ -1010,7 +1054,6 @@ def _constrain_for_sampling(model, max_coefficients, sampling_depth, objective,
         flux_samples = _gapsplit(constrained_model, depth=sampling_depth)
         concordance = _calc_concordance(flux_samples, max_coefficients)
     except:
-        print('WARNING: Flux sampling not completed with given constraints.')
         flux_samples = 'Not performed'
         concordance = 'Not performed'
 
@@ -1042,7 +1085,11 @@ def _calc_concordance(flux_samples, coefficient_dict):
 def _prune_model(model, rm_rxns, conserve):
     
     new_model = deepcopy(model)
-    new_model.id = str(new_model.id) + '_riptide'
+    
+    if str(new_model.id).strip() == '':
+        new_model.id = 'model_riptide'
+    else:
+        new_model.id = str(new_model.id) + '_riptide'
     
     # Parse elements highlighted for pruning based on GPRs
     if conserve == True:
@@ -1290,4 +1337,3 @@ def _maxgap(points, fva):
     target = left + width / 2.0
 
     return relative, target, width
-
